/*P7-1.c*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define IMG					64
#define NUM_B_ARRAY			25
#define B_MIN				200
#define B_STEP				200
#define INITIAL_GUESS_FFAST	0.7
#define INITIAL_GUESS_DFAST	2.0
#define INITIAL_GUESS_FSLOW	0.3
#define INITIAL_GUESS_DSLOW	0.2
#define MASK_THRESHOLD		300.0
#define MAX_LOOP_COUNT		10000
#define MAX_LAMBDA			1.0e20
#define CHI2_GOAL			0.001

void read_data_float(char *fi, float *img, int size);
void write_data_float(char *fo, float *img, int size);
int gauss_jordan(float A[][4], float *B);
int levenberg_marquardt(float *Smeas, float *b_factor, float *a);

int main(void)
{
	float	*DWI;
	int		converge;
	int		j, k;
	float	T2[IMG*IMG];
	float	Smeas[NUM_B_ARRAY], b_factor[NUM_B_ARRAY];
	float	a[4];
	float	ffast[IMG*IMG], Dfast[IMG*IMG], fslow[IMG*IMG], Dslow[IMG*IMG], conductivity[IMG*IMG];
	char	fi[50];

	// 摜t@Cǂݏo߂̃m
	DWI = (float *)malloc(IMG*IMG*NUM_B_ARRAY*sizeof(float));

	// gU摜̃t@Cǂݏo
	for (k = 0; k < NUM_B_ARRAY; k++) {
		sprintf(fi, "b%04d.img", B_MIN+k*B_STEP);
		read_data_float(fi, DWI+k*IMG*IMG, IMG*IMG);

		// bľvZD֋X1000.0Ŋ
		b_factor[k] = (float)(B_MIN+k*B_STEP)/1000.0;
	}

	// T2摜̃t@Cǂݏo
	read_data_float("T2.img", T2, IMG*IMG);

	// sNZɓdvZ
	for (j = 0; j < IMG*IMG; j++) {
		// C̈O邽߁CT2摜ŋPxlMASK_THRESHOLD傫sNẐ݂
		if (T2[j] > MASK_THRESHOLD) {
			// jԖڂ̃sNZɑlo
			for (k = 0; k < NUM_B_ARRAY; k++)
				Smeas[k] = DWI[IMG*IMG*k+j]/T2[j];

			// f֐Ɋ܂܂p[^̏l
			a[0] = INITIAL_GUESS_FFAST;
			a[1] = INITIAL_GUESS_DFAST;
			a[2] = INITIAL_GUESS_FSLOW;
			a[3] = INITIAL_GUESS_DSLOW;

			// f[^_Ƀf֐tBbg
			converge = levenberg_marquardt(Smeas, b_factor, a);

			// a[1]a[3]̑傫fast componentCslow componentƂ
			if (a[1] >= a[3]) {
				ffast[j] = a[0];
				Dfast[j] = a[1]*1.0e-9;
				fslow[j] = a[2];
				Dslow[j] = a[3]*1.0e-9;
			} else {
				ffast[j] = a[2];
				Dfast[j] = a[3]*1.0e-9;
				fslow[j] = a[0];
				Dslow[j] = a[1]*1.0e-9;
			}
		// C̈ł́CSẴp[^[ŏo
		} else {
			ffast[j] = 0.0;
			Dfast[j] = 0.0;
			fslow[j] = 0.0;
			Dslow[j] = 0.0;
		}

		// ďvZ
		conductivity[j] = 2.0*ffast[j]*8.1e8*Dfast[j]/(3.0-ffast[j]);

		// DfastDslow1.0e6āCPʂm2/smm2/sɕϊ
		Dfast[j] *= 1.0e6;
		Dslow[j] *= 1.0e6;
	}

	// 
	free(DWI);

	// ʏo
	write_data_float("ffast.img", ffast, IMG*IMG);
	write_data_float("Dfast.img", Dfast, IMG*IMG);
	write_data_float("fslow.img", fslow, IMG*IMG);
	write_data_float("Dslow.img", Dslow, IMG*IMG);
	write_data_float("conductivity.img", conductivity, IMG*IMG);

	return 0;
}

// 摜t@Cǂݏo
void read_data_float(char *fi, float *img, int size)
{
	FILE *fp;
	
	printf("read data : %s\n", fi);

	if ((fp = fopen(fi, "rb")) == NULL) {
		fprintf(stderr, "Error : file open [%s].\n", fi);
		exit(1);
	}

	fread(img, sizeof(float), size, fp);
	fclose(fp);
}

// ʃt@C̏o
void write_data_float(char *fo, float *img, int size)
{
	FILE *fp;

	printf("write data : %s\n", fo);

	if ((fp = fopen(fo, "wb")) == NULL) {
		fprintf(stderr, "Error : file open [%s].\n", fo);
		exit(1);
	}

	fwrite(img, sizeof(float), size, fp);
	fclose(fp);
}

// _SmeasɊ֐a[0]*exp(-b_factor*a[1])+a[2]*exp(-b_factor*a[3])tBbg
int levenberg_marquardt(float *Smeas, float *b_factor, float *a)
{
	float	Scalc_old[NUM_B_ARRAY], Scalc_new[NUM_B_ARRAY];
	float	dS_da_old[4][NUM_B_ARRAY], dS_da_new[4][NUM_B_ARRAY];
	float	A[4][4], B[4];
	float	chi2_old = 0.0, chi2_new;
	float	lambda = 0.001;
	int		regular_matrix;
	int		i, j, k;
	int		loop_count = 0;

	// e_ɂāCfvZMxScalc_oldCMx̔dS_da_oldCJC2chi2_oldvZ
	for (k = 0; k < NUM_B_ARRAY; k++) {
		dS_da_old[0][k] = exp(-b_factor[k]*a[1]);
		dS_da_old[1][k] = -b_factor[k]*a[0]*dS_da_old[0][k];
		dS_da_old[2][k] = exp(-b_factor[k]*a[3]);
		dS_da_old[3][k] = -b_factor[k]*a[2]*dS_da_old[2][k];
		Scalc_old[k] = a[0]*dS_da_old[0][k]+a[2]*dS_da_old[2][k];
		chi2_old += (Smeas[k]-Scalc_old[k])*(Smeas[k]-Scalc_old[k]);
	}

	// 񐔂MAX_LOOP_COUNT𒴂ClambdaMAX_LAMBDA𒴂Ȃ͈͂ŌJԂ
	while (loop_count < MAX_LOOP_COUNT && lambda < MAX_LAMBDA) {
		// ȗs(Hesses)̊evfߎvZ
		for (i = 0; i < 4; i++)
			for (j = 0; j < 4; j++) {
				if (j >= i) {
					A[i][j] = 0.0;
					for (k = 0; k < NUM_B_ARRAY; k++)
						A[i][j] += dS_da_old[i][k]*dS_da_old[j][k];
					if (j == i)
						// Ίpvf(1+lambda){
						A[i][j] *= (1+lambda);
				} else {
					// ȗs̑Ώ̐𗘗pāCΊpvf荶Ɉʒuvf̌vZȗ
					A[i][j] = A[j][i];
				}
			}

		// JC2̌zxNg̊evZ
		for (i = 0; i < 4; i++) {
			B[i] = 0.0;
			for (k = 0; k < NUM_B_ARRAY; k++)
				B[i] += (Smeas[k]-Scalc_old[k])*dS_da_old[i][k];
		}

		// sȃBvZ
		regular_matrix = gauss_jordan(A, B);

		// XVsa + BɂāCJC2chi2_newȂǂvZ
		chi2_new = 0.0;
		for (k = 0; k < NUM_B_ARRAY; k++) {
			dS_da_new[0][k] = exp(-b_factor[k]*(a[1]+B[1]));
			dS_da_new[2][k] = exp(-b_factor[k]*(a[3]+B[3]));
			Scalc_new[k] = (a[0]+B[0])*dS_da_new[0][k]+(a[2]+B[2])*dS_da_new[2][k];
			chi2_new += (Smeas[k]-Scalc_new[k])*(Smeas[k]-Scalc_new[k]);
		}

		// B߂Ȃ΁C܂́Cp[^̍XVɂăJC2悪Ȃ΁Clambda10{
		if (chi2_new >= chi2_old || regular_matrix == 0)
			lambda *= 10.0;
		// JC2悪ꍇ́CXVs̗p
		else {
			for (i = 0; i < 4; i++)
				a[i] += B[i];

			// XVsɂāCMx̔ȂǂvZ
			for (k = 0; k < NUM_B_ARRAY; k++) {
				dS_da_old[0][k] = dS_da_new[0][k];
				dS_da_old[1][k] = -b_factor[k]*a[0]*dS_da_old[0][k];
				dS_da_old[2][k] = dS_da_new[2][k];
				dS_da_old[3][k] = -b_factor[k]*a[2]*dS_da_old[2][k];
				Scalc_old[k] = Scalc_new[k];
			}

			// JC2̕ωCHI2_GOALCƔfČJԂID߂l1
			if (chi2_old-chi2_new < chi2_old*CHI2_GOAL)
				return 1;

			// lambda0.1{
			chi2_old = chi2_new;
			lambda *= 0.1;
		}

		loop_count++;
	}

	// Ȃ΁C߂l0
	return 0;
}

// 4ϐ̘AꎟDA͌WsCB͒萔xNg
int gauss_jordan(float A[][4], float *B)
{
	int		i, j, k;
	int		pivot;
	float	largest, temp;

	// ł̗(i = 0)珇ɁC|os
	for (i = 0; i < 4; i++) {
		largest = 0.0;
		// isڂ牺ɁCiڂTāCő̐ΒlvfD̍sԍpivotɑ
		for (k = i; k < 4; k++)
			if (fabs(A[k][i]) >= largest) {
				largest = fabs(A[k][i]);
				pivot = k;
			}

		// pivotivȂ΁CpivotsڂisڂւD͌vZ덷}邽
		if (pivot != i) {
			for (j = i; j < 4; j++) {
				temp = A[i][j];
				A[i][j] = A[pivot][j];
				A[pivot][j] = temp;
			}

			temp = B[i];
			B[i] = B[pivot];
			B[pivot] = temp;
		}

		// ȉCisivfpivotƂđ|os
		pivot = i;

		// pivotvf[ȂCWs񂪓ٍsłCȂD߂l0
		if (A[pivot][pivot] == 0.0)
			return 0;
		else {
			// pivotsڂ̊evfA[pivot][pivot]Ŋ
			for (j = pivot+1; j < 4; j++)
				A[pivot][j] /= A[pivot][pivot];
			B[pivot] /= A[pivot][pivot];

			// k(!=pivot)sڂCpivotsڂ萔{̂
			for (k = 0; k < 4; k++)
				if (k != pivot) {
					for (j = pivot+1; j < 4; j++)
						A[k][j] -= A[pivot][j]*A[k][pivot];
					B[k] -= B[pivot]*A[k][pivot];
				}
		}
	}
	// ꍇC߂l1
	return 1;
}