/*----------------------------------------------------
	
	P8-8 relaxation_runge.cpp

	(86(3) ɑΉ)
	QNb^@pɘa摜̍쐬

----------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PN	10			/* number of parameters + 1 */
#define PI	3.14159265358979

typedef struct {
	char	f1[50];		/* input T1 image file name */
	char	f2[50];		/* input T2 image file name */
	char	f3[50];		/* input P0 image file name */
	char	f4[50];		/* output longitudinal relaxation image file name */
	char	f5[50];		/* output transverse relaxation image file name */
	float	*t1;			/* real image */
	float	*t2;			/* imaginary image */
	float	*p0;			/* spectrum */
	float	*im_mz;		/* imaginary image */
	float	*im_mxy;	/* spectrum */
	double	t;			/* relaxation time */
	double	h;			/* step width */
	int		nx;				/* number of matrix (x) */
	int		ny;				/* number of matrix (y) */
} Param;

char *menu[PN] = {
	"P8-8 relaxation_runge.cpp",
	"Input T1 image file name             <float> ",
	"Input T2 image file name             <float> ",
	"Input P0 image file name             <float> ",
	"    Relaxation time      (t)   ",
	"    Step width           (h)   ",
	"    Number of matrix     (x)   ",
	"    Number of matrix     (y)   ",
	"Output T1 relaxation image file name <float> ",
	"Output T2 relaxation image file name <float> ",
};

void relaxation(float *, float *, float *, float *, float *, double , double , int , int);
double mz(float , float , double , double);
double mxy(float , float , double , double);
double cal_mz(double , double);
double cal_mxy(double , double);
void read_data(char *, float *, int);
void write_data(char *, float *, int);
void write_data_ascii(char *, float *, int);
void img_default(float *, float , int);

void getparameter(Param *pm)
{
	int		i;
	char	dat[256];

	/* default parameter value */
	sprintf(pm->f1, "c_t1v_sa60.img");
	sprintf(pm->f2, "c_t2v_sa60.img");
	sprintf(pm->f3, "c_p0v_sa60.img");
	sprintf(pm->f4, "tate.img");
	sprintf(pm->f5, "yoko.img");
	pm->t  = 20.0;
	pm->h  = 1.0;
	pm->nx = 256;
	pm->ny = 256;

	i = 0;
	fprintf(stdout, "\n%s\n\n", menu[i++]);
	fprintf(stdout, " %s [%s] :", menu[i++], pm->f1);
	if (*gets(dat) != '\0') strcpy(pm->f1, dat);
	fprintf(stdout, " %s [%s] :", menu[i++], pm->f2);
	if (*gets(dat) != '\0') strcpy(pm->f2, dat);
	fprintf(stdout, " %s [%s] :", menu[i++], pm->f3);
	if (*gets(dat) != '\0') strcpy(pm->f3, dat);
	fprintf(stdout, " %s [%.1f] :", menu[i++], pm->t);
	if (*gets(dat) != '\0') pm->t = atof(dat);
	fprintf(stdout, " %s [%.1f] :", menu[i++], pm->h);
	if (*gets(dat) != '\0') pm->h = atof(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], pm->nx);
	if (*gets(dat) != '\0') pm->nx = atoi(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], pm->ny);
	if (*gets(dat) != '\0') pm->ny = atoi(dat);
	fprintf(stdout, " %s [%s] :", menu[i++], pm->f4);
	if (*gets(dat) != '\0') strcpy(pm->f4, dat);
	fprintf(stdout, " %s [%s] :", menu[i++], pm->f5);
	if (*gets(dat) != '\0') strcpy(pm->f5, dat);
}

int main(void)
{
	Param	*pm;
	pm = (Param *)malloc(sizeof(Param));
	getparameter(pm);

	pm->t1 = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->t2 = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->p0 = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	pm->im_mz = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));
	pm->im_mxy = (float *)malloc((unsigned long)pm->nx*pm->ny*sizeof(float));

	printf("\n T1l摜̓ \n");
    read_data(pm->f1, pm->t1, pm->nx*pm->ny);

	printf(" T2l摜̓ \n");
    read_data(pm->f2, pm->t2, pm->nx*pm->ny);

	printf(" P0l摜̓ \n");
    read_data(pm->f3, pm->p0, pm->nx*pm->ny);

	printf(" ɘǎvZ \n");
	relaxation(pm->im_mz, pm->im_mxy, pm->t1, pm->t2, pm->p0, pm->t, pm->h, pm->nx, pm->ny);

	printf(" cɘa摜̏o \n");
    write_data(pm->f4, pm->im_mz, pm->nx*pm->ny);

	printf(" ɘa摜̏o \n");
    write_data(pm->f5, pm->im_mxy, pm->nx*pm->ny);
	
	free(pm->t1);
	free(pm->t2);
	free(pm->p0);
	free(pm->im_mz);
	free(pm->im_mxy);

	return 0;
}

void relaxation(float *im_mz, float *im_mxy, float *t1, float *t2, float *p0, double t, double h, int nx, int ny)
// esNZƂɊɘavZ
{
	int		i;

	// cɘǎvZ
	for (i = 0; i < nx*ny; i++) {
		if (t1[i] == 0.0) im_mz[i] = 0.0;
		else im_mz[i] = (float)mz(t1[i], p0[i], t, h);
	}

	// ɘǎvZ
	for (i = 0; i < nx*ny; i++) {
		if (t2[i] == 0.0) im_mxy[i] = 0.0;
		else im_mxy[i] = (float)mxy(t2[i], p0[i], t, h);
	}
}

double mz(float t1, float p0, double t, double h)
// cɘa(longitudinal relaxation)̌vZ
{
	int		i, n;
	double	x;
	double	k1, k2, k3, k4;
	
	n  = (int)t/h; 
	x  = 0.0;
	k1 = k2 = k3 = k4 = 0.0;

	for (i = 0; i <= n; i++) {
        k1 = cal_mz(x         , (double)t1);
		k2 = cal_mz(x+h*k1/2.0, (double)t1);
		k3 = cal_mz(x+h*k2/2.0, (double)t1);
		k4 = cal_mz(x+h*k3    , (double)t1);
       
		x += h*(k1+2.0*k2+2.0*k3+k4)/6.0;
	}
	x *= p0;

	return x;
}

double mxy(float t2, float p0, double t, double h)
// ɘa(transverse relaxation)̌vZ
{
	int		i, n;
	double	x;
	double	k1, k2, k3, k4;
	
	n  = (int)t/h;
	x  = 1.0;
	k1 = k2 = k3 = k4 = 0.0;

	for (i = 0; i <= n; i++) {
        k1 = cal_mxy(x         , (double)t2);
		k2 = cal_mxy(x+h*k1/2.0, (double)t2);
		k3 = cal_mxy(x+h*k2/2.0, (double)t2);
		k4 = cal_mxy(x+h*k3    , (double)t2);
		x += h*(k1+2.0*k2+2.0*k3+k4)/6.0;
	}
	x *= p0;

	return x;
}

double cal_mz(double x, double T1)
// cɘa(longitudinal relaxation)̔̌vZ
{
	return (1.0-x)/T1;		
}

double cal_mxy(double x, double T2)
// ɘa(transverse relaxation)̔̌vZ
{
	return -x/T2;		
}
				  
/*===================
	routine
===================*/
void read_data(char *fi, float *prj, int size)
{
	FILE	*fp;

	/* open file and read data */
	if ((fp = fopen(fi, "rb")) == NULL) {
		fprintf(stderr, " Error : file open [%s].\n", fi);
		exit(1);
	}
	fread(prj, sizeof(float), size, fp);
	fclose(fp);
}

void write_data(char *fi, float *img, int size)
{
	FILE	*fp;

	/* open file and write data */
	if ((fp = fopen(fi, "wb")) == NULL) {
		fprintf(stderr, " Error : file open [%s].\n", fi);
		exit(1);
	}
	fwrite(img, sizeof(float), size, fp);
	fclose(fp);
}

void write_data_ascii(char *fi, float *prj, int size)
{
	int		i;
	FILE	*fp;

	/* open file and write data */
	if ((fp = fopen(fi, "w")) == NULL) {
		fprintf(stderr, " Error : file open [%s].\n", fi);
		exit(1);
	}
	for (i = 0 ; i < size ; i++) {
		fprintf(fp, "%f\n", prj[i]);
	}
	fclose(fp);
}

void img_default(float *im, float value, int pix)
{
	int		i;

	for (i = 0; i < pix; i++)
		im[i] = value;
}	